import os
import sys
import asyncio
import aiohttp

from ipaddress import ip_address as address
from urllib.parse import urlparse

from .config import PAWSCONFIG, VERSION, script_path, logging
from . import middleware


def webserver():
	from . import views

	web = aiohttp.web.Application(middlewares=[
		middleware.http_filter,
		middleware.http_signatures,
		middleware.http_redirect
	])

	web.add_routes([
		aiohttp.web.route('GET', '/authorize', views.authorize),
	])

	return web


async def start_webserver():
	app = webserver()
	runner = aiohttp.web.AppRunner(app, access_log_format='%{X-Real-Ip}i "%r" %s %b "%{User-Agent}i"')
	await runner.setup()

	listen = PAWSCONFIG['host']
	port = PAWSCONFIG['port']

	if listen.startswith('unix:'):
		if sys.platform != 'win32':
			sock_listen = listen.replace('unix:', '')

			logging.info(f'Starting webserver at socket: {sock_listen}')
			site = aiohttp.web.UnixSite(runner, sock_listen)

		else:
			logging.error('Windows cannot use unix sockets. Use an IP address instead. Exiting...')
			sys.exit()

	else:
		try:
			address(listen)

		except ValueError:
			logging.warning('Invalid IP address. Listening on "0.0.0.0" instead.')
			listen = '127.0.0.1'

		try:
			int(port)

		except ValueError:
			logging.warning('Invalid port. Using 3621 instead.')
			port = 3621

		logging.info(f'Starting webserver at address: {listen}:{port}')
		site = aiohttp.web.TCPSite(runner, listen, port)

	await site.start()


def main():
	try:
		loop = asyncio.get_event_loop()
		asyncio.ensure_future(start_webserver())
		loop.run_forever()

	except KeyboardInterrupt:
		logging.info('Bye!')
